#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "../SimionPA/SimionPA.h"

#ifdef PAUSE
#include <unistd.h>
#endif

void display_grid(PAHeader head, double ***g){
        int y = 0;
        int x = 0;
        int z = 0;

        printf("-----------------------------------\n");
        for (z=0; z < head.nz; z++) {
                printf("{\n");
                for (y=0; y < head.ny; y++) {
                        printf("\t{");
                        for (x=0; x<head.nx; x++) {
                                char suf[2];
                                if (!(x == (head.nx - 1))) {
                                        strcpy(suf,",");
                                }else{
                                        strcpy(suf,"");
                                }

                                if (g[z][y][x] >= head.max_voltage) {
                                        printf("%+2.3lf%s", g[z][y][x] - 2 * head.max_voltage , suf);
                                } else {
                                        printf("%+2.3lf%s", g[z][y][x], suf);
                                }

                        }
                        if (y == (head.ny - 1)){
                                printf("}\n");
                                continue;
                        }
                        printf("},\n");
                }
                if (z == (head.nz - 1)){
                        printf("}\n");
                        continue;
                }
                printf("},\n");
        }
        printf("-----------------------------------\n");
}

int is_electrode(PAHeader head, double point) {
	return point > head.max_voltage;
}

double average_potential(PAHeader head, double ***g) {
	int x, y, z;
	double average = 0;
	int electrode_count = 0;

	for (z = 0; z < head.nz; z++) {
		for (y = 0; y < head.ny; y++) {
			for (x = 0; x < head.nx; x++) {
				if (is_electrode(head, g[z][y][x])) {
					average += (g[z][y][x] - 2 * head.max_voltage);
					electrode_count++;
				}
			}
		}
	}

	average = average / electrode_count;
	printf("%d electrode points found, with an average potential of %lf\n", electrode_count, average);
	return average;
}

void flood_grid(PAHeader head, double ***g, double value) {
	int x,y,z;
		
	for (z = 0; z < head.nz; z++) {
		for (y = 0; y < head.ny; y++) {
			for (x = 0; x < head.nx; x++) {
				if (!is_electrode(head, g[z][y][x])){
					g[z][y][x] = value;
				}
			}
		}
	}
	printf("Non-electrode points set to %lf\n", value);
}

double iterate(PAHeader head, double ***g, double orf) {
	// x, y and z are coordinates, i is used to keep track of the number of points used for this average
	// Pn convention matches that used in section H.3 of the SIMION manual
	int x = 0; 
	int y = 0;
	int z = 0;
	int i = 0;
	
	double avg_val = 0;
	double max_change = 0;

	for (z = 0; z < head.nz; z++) {
		for (y = 0; y < head.ny; y++) {
			for (x = 0; x < head.nx; x++) {
				if ( g[z][y][x] > head.max_voltage ) { 
					#ifdef DEBUG2
					printf("Skipping point (%d, %d, %d) - Electrode\n", x, y, z);
					#endif
					continue; 
				}
				#ifdef DEBUG2
				printf("Processing point (%d, %d, %d)\n", x, y, z);	
				#endif
				avg_val = 0;
				i = 0;
				if (head.nx > 1) {
					if (x == 0 && head.nx > 1) {
						#ifdef DEBUG3
						printf ("P0: (%d, %d, %d)\tP1: N/A  \tP2: (%d, %d, %d)\t\t", x, y, z, 1, y, z);
						#endif
						avg_val += ( g[z][y][1] > head.max_voltage ? g[z][y][1] - 2 * head.max_voltage : g[z][y][1] ); //P2
						i++;
						if (head.mirror & SIMION_MIRROR_X) { // Mirrored in X, so P2 also appears on the other side of the mirror
							avg_val += ( g[z][y][1] > head.max_voltage ? g[z][y][1] - 2 * head.max_voltage : g[z][y][1] ); //P2
							i++;
						}
					} else if (x == head.nx - 1 && (head.nx -1) > 0) {
						#ifdef DEBUG3
						printf ("P0: (%d, %d, %d)\tP1: (%d, %d, %d) \tP2: N/A  \t\t", x, y, z, x-1, y, z);
						#endif
						avg_val += (g[z][y][x-1] > head.max_voltage ? g[z][y][x-1] - 2 * head.max_voltage : g[z][y][x-1] ); //P1
						i++;
					} else {
						#ifdef DEBUG3
						printf ("P0: (%d, %d, %d)\tP1: (%d, %d, %d) \tP2: (%d, %d, %d)\t\t", x, y, z, x-1, y, z, x+1, y, z);
						#endif
						avg_val += (g[z][y][x-1] > head.max_voltage ? g[z][y][x-1] - 2 * head.max_voltage : g[z][y][x-1] ); //P1
						avg_val += (g[z][y][x+1] > head.max_voltage ? g[z][y][x+1] - 2 * head.max_voltage : g[z][y][x+1] ); //P2
						i += 2;
					}
				}

				if (head.ny > 1) {
					if (y == 0 && head.ny > 1) {
						#ifdef DEBUG3
						printf ("P3: (%d, %d, %d) \tP4: N/A\t\t", x, y+1, z);
						#endif
						avg_val += (g[z][1][x] > head.max_voltage ? g[z][1][x] - 2 * head.max_voltage : g[z][1][x] ); //P3
						i++;
						if (head.mirror & SIMION_MIRROR_Y) { //Mirrored in Y
							avg_val += (g[z][1][x] > head.max_voltage ? g[z][1][x] - 2 * head.max_voltage : g[z][1][x] ); //P3
							i++;
						}
					} else if (y == head.ny - 1 && (head.ny - 1) > 0 ) {
						#ifdef DEBUG3
						printf ("P3: N/A  \tP4: (%d, %d, %d)\t\t", x, y-1, z);
						#endif
						avg_val += (g[z][(y-1)][x] > head.max_voltage ? g[z][(y-1)][x] - 2 * head.max_voltage : g[z][(y-1)][x] ); //P4
						i++;
					} else {
						#ifdef DEBUG3
						printf ("P3: (%d, %d, %d) \tP4: (%d, %d, %d)\t\t", x, y+1, z, x, y-1, z);
						#endif
						avg_val += (g[z][(y+1)][x] > head.max_voltage ? g[z][(y+1)][x] - 2 * head.max_voltage : g[z][(y+1)][x] ); //P3
						avg_val += (g[z][(y-1)][x] > head.max_voltage ? g[z][(y-1)][x] - 2 * head.max_voltage : g[z][(y-1)][x] ); //P4
						i += 2;
					}
				}

				if (head.nz > 1) {
					if (z == 0) {
						#ifdef DEBUG3
						printf ("P5: (%d, %d, %d) \tP6: N/A\n", x, y, 1);
						#endif
						avg_val += (g[1][y][x] > head.max_voltage ? g[1][y][x] - 2 * head.max_voltage : g[1][y][x] ); //P5
						i++;
						if (head.mirror & SIMION_MIRROR_Z) { //Mirrored in Z
							avg_val += (g[1][y][x] > head.max_voltage ? g[1][y][x] - 2 * head.max_voltage : g[1][y][x] ); //P5
							i++;
						}
					} else if (z == head.nz - 1) {
						#ifdef DEBUG3
						printf ("P5: N/A  \tP6: (%d, %d, %d)\n", x, y, z-1);
						#endif
						avg_val += (g[z-1][y][x] > head.max_voltage ? g[z-1][y][x] - 2 * head.max_voltage : g[z-1][y][x] ); //P6
						i++;
					} else {
						#ifdef DEBUG3
						printf ("P5: (%d, %d, %d) \tP6: (%d, %d, %d)\n", x, y, z+1, x, y, z-1);
						#endif
						avg_val += (g[z+1][y][x] > head.max_voltage ? g[z+1][y][x] - 2 * head.max_voltage : g[z+1][y][x] ); //P5
						avg_val += (g[z-1][y][x] > head.max_voltage ? g[z-1][y][x] - 2 * head.max_voltage : g[z-1][y][x] ); //P6
						i += 2;
					}
				}
	                        avg_val = avg_val / i;
	                        double change = avg_val - g[z][y][x];
	                        double new_val = g[z][y][x]+(change) * (1 + orf);
	
	                        #ifdef DEBUG2
	                        printf("\nPoint (%d, %d):\n\tOld value: \t%+.8lf\n\tAverages: \t%.8lf\n", x, y, g[z][y][x], avg_val);
	                        printf("\tChange: \t%.8lf\n\tNew value: \t%.8lf\n\ti:\t\t%d\n", change, new_val, i);
	                        #endif
	
	                        if (fabs(change) > max_change) {
	                                max_change = fabs(change);
	                        }
				#ifdef PAUSE
				usleep(PAUSE * 1000);
				#endif
	                        g[z][y][x] = new_val;
		
			}
		}
	}
	#ifdef DEBUG	
	printf("Max change: %.8lf\n", max_change);
	#endif
	return max_change;
}

double iterate_cylindrical(PAHeader head, double ***g, double orf) {
	// Pn convention matches that used in section H.3 of the SIMION manual
	// Cylindrical symmetry doesn't allow for the nice shortcuts used above, so requires it's own routine
	// Axis of rotation is always y=0
	// x mirroring is permitted, z mirroring is forbidden
	// head.nz == 1

	int x = 0; 
	int j = 0;
	int y = 0;
	
	double avg_val = 0;
	double max_change = 0;

	for (j = -(head.ny - 1); j < head.ny; j++) {
		for (x = 0; x < head.nx; x++) {
			y=abs(j);
			if ( g[0][y][x] > head.max_voltage ) { 
				#ifdef DEBUG2
				printf("Skipping point (%d, %d) - Electrode\n", x, y);
				#endif
				continue; 
			}
			#ifdef DEBUG2
			printf("Processing point (%d, %d)\n", x, y);	
			#endif
			avg_val = 0;

			if (!(head.nx > 1 && head.ny > 1)) { return -1;}

			if (x == 0 && y == 0) { //LL Corner (Corner Axis point)
				#ifdef DEBUG3
				printf ("P0: (%d, %d)\tP1: N/A  \tP2: (%d, %d)\t\tP3: (%d, %d)\tP4: N/A\n", x, y, 1, y, x, y+1);
				#endif
				double P2 = (g[0][y][x+1] > head.max_voltage ? g[0][y][x+1] - 2 * head.max_voltage : g[0][y][x+1] ); // P2
                                double P3 = (g[0][(y+1)][x] > head.max_voltage ? g[0][(y+1)][x] - 2 * head.max_voltage : g[0][(y+1)][x] ); //P3
				avg_val = P2 / 5 + 4 * P3 / 5;
			} else if (x == head.nx-1 && y == 0) {//LR Corner (Corner Axis point)
				#ifdef DEBUG3
				printf ("P0: (%d, %d)\tP1: (%d, %d)\tP2: N/A   \t\tP3: (%d, %d)\tP4: N/A\n", x, y, x-1, y, x, y+1);
				#endif
                                double P1 = (g[0][y][x-1] > head.max_voltage ? g[0][y][x-1] - 2 * head.max_voltage : g[0][y][x-1] ); // P1
                                double P3 = (g[0][(y+1)][x] > head.max_voltage ? g[0][(y+1)][x] - 2 * head.max_voltage : g[0][(y+1)][x] ); //P3
				avg_val = P1 / 5 + 4 * P3 / 5;
			} else if (x == head.nx-1 && y == head.ny -1) { //UR Corner (Upper corner point)
				#ifdef DEBUG3
				printf ("P0: (%d, %d)\tP1: (%d, %d)\tP2: N/A  \t\tP3: N/A   \tP4: (%d, %d)\n", x, y, x-1, y, x, y-1);
				#endif
                                double P1 = (g[0][y][x-1] > head.max_voltage ? g[0][y][x-1] - 2 * head.max_voltage : g[0][y][x-1] ); // P1
                                double P4 = (g[0][(y-1)][x] > head.max_voltage ? g[0][(y-1)][x] - 2 * head.max_voltage : g[0][(y-1)][x] ); //P4				
				avg_val = (2 * y * P1 + (2*y-1) * P4 )/ (4*y-1);
			} else if (x == 0  && y == head.ny -1) { // UL Corner (Upper corner point)
				#ifdef DEBUG3
				printf ("P0: (%d, %d)\tP1: N/A  \tP2: (%d, %d)\t\tP3: N/A  \tP4: (%d, %d)\n", x, y, x+1, y, x, y-1);
				#endif
				double P2 = (g[0][y][x+1] > head.max_voltage ? g[0][y][x+1] - 2 * head.max_voltage : g[0][y][x+1] ); // P2
                                double P4 = (g[0][(y-1)][x] > head.max_voltage ? g[0][(y-1)][x] - 2 * head.max_voltage : g[0][(y-1)][x] ); //P4				
				avg_val = (2 * y * P2 + (2*y-1) * P4 )/ (4*y-1);
			} else if (y == 0) { //Interior axis points
				#ifdef DEBUG3
				printf ("P0: (%d, %d)\tP1: (%d, %d)\tP2: (%d, %d)\t\tP3: (%d, %d)\tP4: N/A\n", x, y, x-1, y, x+1, y, x, y+1);
				#endif
                                double P1 = (g[0][y][x-1] > head.max_voltage ? g[0][y][x-1] - 2 * head.max_voltage : g[0][y][x-1] ); // P1
				double P2 = (g[0][y][x+1] > head.max_voltage ? g[0][y][x+1] - 2 * head.max_voltage : g[0][y][x+1] ); // P2
                                double P3 = (g[0][(y+1)][x] > head.max_voltage ? g[0][(y+1)][x] - 2 * head.max_voltage : g[0][(y+1)][x] ); //P3	
				avg_val = (P1 + P2) / 6 + 2 * P3 / 3;
			} else if (y == head.ny-1) { //Top Edge points
				#ifdef DEBUG3
				printf ("P0: (%d, %d)\tP1: (%d, %d)\tP2: (%d, %d)\t\tP3: N/A   \tP4: (%d, %d)\n", x, y, x-1, y, x+1, y, x, y-1);
				#endif
                                double P1 = (g[0][y][x-1] > head.max_voltage ? g[0][y][x-1] - 2 * head.max_voltage : g[0][y][x-1] ); // P1
				double P2 = (g[0][y][x+1] > head.max_voltage ? g[0][y][x+1] - 2 * head.max_voltage : g[0][y][x+1] ); // P2
                                double P4 = (g[0][(y-1)][x] > head.max_voltage ? g[0][(y-1)][x] - 2 * head.max_voltage : g[0][(y-1)][x] ); //P4	
				avg_val = (2 * y *(P1 + P2 + P4) - P4) / (6*y-1);
			} else if (x == 0) { //Left edge points
				#ifdef DEBUG3
				printf ("P0: (%d, %d)\tP1: N/A  \tP2: (%d, %d)\t\tP3: (%d, %d)\tP4: (%d, %d)\n", x, y, x+1, y, x, y+1, x, y-1);
				#endif
				double P2 = (g[0][y][x+1] > head.max_voltage ? g[0][y][x+1] - 2 * head.max_voltage : g[0][y][x+1] ); // P2
                                double P3 = (g[0][(y+1)][x] > head.max_voltage ? g[0][(y+1)][x] - 2 * head.max_voltage : g[0][(y+1)][x] ); //P3
                                double P4 = (g[0][(y-1)][x] > head.max_voltage ? g[0][(y-1)][x] - 2 * head.max_voltage : g[0][(y-1)][x] ); //P4	
				double PA = P3 + ((2 + (1.0 / y) ) / 4.0)*(P4-P3);
				avg_val = (P2 + 2 * PA) / 3;
			} else if (x == head.nx-1) { //Right edge points
				#ifdef DEBUG3
				printf ("P0: (%d, %d)\tP1: (%d, %d)\tP2: N/A  \t\tP3: (%d, %d)\tP4: (%d, %d)\n", x, y, x-1, y, x, y+1, x, y-1);
				#endif
                                double P1 = (g[0][y][x-1] > head.max_voltage ? g[0][y][x-1] - 2 * head.max_voltage : g[0][y][x-1] ); // P1
                                double P3 = (g[0][(y+1)][x] > head.max_voltage ? g[0][(y+1)][x] - 2 * head.max_voltage : g[0][(y+1)][x] ); //P3
                                double P4 = (g[0][(y-1)][x] > head.max_voltage ? g[0][(y-1)][x] - 2 * head.max_voltage : g[0][(y-1)][x] ); //P4	
				double PA = P3 + ((2 + (1.0 / y) ) / 4.0)*(P4-P3);
				avg_val = (P1 + 2 * PA) / 3;
			} else {
				#ifdef DEBUG3
				printf ("P0: (%d, %d)\tP1: (%d, %d)\tP2: (%d, %d)\t\tP3: (%d, %d)\tP4: (%d, %d)\n", x, y, x-1, y, x+1, y, x, y+1, x, y-1);
				#endif
                                double P1 = (g[0][y][x-1] > head.max_voltage ? g[0][y][x-1] - 2 * head.max_voltage : g[0][y][x-1] ); // P1
				double P2 = (g[0][y][x+1] > head.max_voltage ? g[0][y][x+1] - 2 * head.max_voltage : g[0][y][x+1] ); // P2
                                double P3 = (g[0][(y+1)][x] > head.max_voltage ? g[0][(y+1)][x] - 2 * head.max_voltage : g[0][(y+1)][x] ); //P3
                                double P4 = (g[0][(y-1)][x] > head.max_voltage ? g[0][(y-1)][x] - 2 * head.max_voltage : g[0][(y-1)][x] ); //P4	
				double PA = P3 + ((2 + (1.0 / y) ) / 4.0)*(P4-P3);
				avg_val = (P1 + P2 + 2 * PA ) / 4;
			}
				
			
			double change = avg_val - g[0][y][x];
			double new_val = g[0][y][x]+(change) * (1 + orf);

			#ifdef DEBUG2
			printf("\nPoint (%d, %d):\n\tOld value: \t%+.8lf\n\tAverages: \t%.8lf\n", x, y, g[0][y][x], avg_val);
			printf("\tChange: \t%.8lf\n\tNew value: \t%.8lf\n", change, new_val);
			#endif

			if (fabs(change) > max_change) {
				max_change = fabs(change);
			}
			#ifdef PAUSE
			usleep(PAUSE * 1000);
			#endif
			g[0][y][x] = new_val;
	
		}
	}
	#ifdef DEBUG	
	printf("Max change: %.8lf\n", max_change);
	#endif
	return max_change;
}

int main(int argc, char *argv[]){
	printf("Dynamic Over-Relaxation test\n\nIterate over a potential array (from file)  and print values for each iteration\nUses the dynamic over-relaxation factor method laid out in H.5 of the SIMION Manual\n");

	double ***grid = NULL;
	
	PAHeader pa_head;

	int iterations = 0;
	double orf = 0.40;
	double orf_max = 0.90;
	double hist = 0.70;
	double goal = 0.005;
	double del = 0;

	if (argc == 2) {
		if ((grid = load_pa_file(argv[1], &pa_head, grid)) == NULL){
			perror("Error loading file");
			return EXIT_FAILURE;
		}
	} else if (argc == 3) {
		#ifdef DEBUG
		printf("Goal specified: %f\n", goal);
		#endif
		if (!(goal = atof(argv[2])) > 0) {
			fprintf(stderr, "Goal must be a positive number - not %+f\n", goal);
			return EXIT_FAILURE;
		}
		if ((grid = load_pa_file(argv[1], &pa_head, grid)) == NULL){
			perror("Error loading file");
			return EXIT_FAILURE;
		}
	} else {
		fprintf(stderr, "Usage: %s filename [goal]\nDefault goal is %f\n", argv[0], goal);
		return EXIT_FAILURE;
	}

	//display_grid(pa_head, grid);
	/* if (pa_head.symmetry == SIMION_CYLINDRICAL_SYMMETRY) {
		printf("Getting average potential...\n");
		double avg_val = average_potential(pa_head, grid);
		printf("Setting non-electrode points to %lf...\n", -5*avg_val);
		flood_grid(pa_head, grid, -5*avg_val);
	} */
	while ( !(iterations > 1000000) ) {
		if (pa_head.symmetry == SIMION_PLANAR_SYMMETRY) {
			if ((del=iterate(pa_head, grid, orf)) <= goal ) {break;}
		} else if (pa_head.symmetry == SIMION_CYLINDRICAL_SYMMETRY) {
			if ((del=iterate_cylindrical(pa_head, grid, orf)) <= goal)  {break;}
		}
		iterations++;
		#ifdef DEBUG
		printf("Iteration: %d\nORF: %.8lf\n", iterations, orf);
		#endif
		if (iterations <= 10) {
			orf = 0.67;
		} else {
			orf = (orf * hist) + (1 - hist)*(orf_max - (orf_max-0.4)*(1.0 / ( 1 + (del / (2 * goal) ) ) ) );  	
		}

	}
	printf("\nSolution after %d iterations:\n", iterations);
	display_grid(pa_head, grid);
	free_points(pa_head, grid);
	return EXIT_SUCCESS;
}
