#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "../SimionPA/SimionPA.h"

#ifdef PAUSE
#include <unistd.h>
#endif

const double unit_charge = 1; //0.0000000000000000001602176565; //1.6 * pow(10, -19);
const double e_mass = 511000;//0.0000000000000000000000000000009109382; //9.11 * pow(10, -31);
double little_omega = 0; //Initialise these in main() 
double BigOmega = 0;
const int n0 = 1000;
const double centralV = 10;
const double endcapV = 80;
double T = 0; //Initialise in main

void display_longgrid(PAHeader head, long double ***g){
        int y = 0;
        int x = 0;
        int z = 0;

        printf("-----------------------------------\n");
        for (z=0; z < head.nz; z++) {
                printf("{\n");
                for (y=0; y < head.ny; y++) {
                        printf("\t{");
                        for (x=0; x<head.nx; x++) {
                                char suf[2];
                                if (!(x == (head.nx - 1))) {
                                        strcpy(suf,",");
                                }else{
                                        strcpy(suf,"");
                                }

                                if (g[z][y][x] >= head.max_voltage) {
                                        printf("%+2.3Lf%s", g[z][y][x] - 2 * head.max_voltage , suf);
                                } else {
                                        printf("%+2.3Lf%s", g[z][y][x], suf);
                                }

                        }
                        if (y == (head.ny - 1)){
                                printf("}\n");
                                continue;
                        }
                        printf("},\n");
                }
                if (z == (head.nz - 1)){
                        printf("}\n");
                        continue;
                }
                printf("},\n");
        }
        printf("-----------------------------------\n");
}
void display_grid(PAHeader head, double ***g){
        int y = 0;
        int x = 0;
        int z = 0;

        printf("-----------------------------------\n");
        for (z=0; z < head.nz; z++) {
                printf("{\n");
                for (y=0; y < head.ny; y++) {
                        printf("\t{");
                        for (x=0; x<head.nx; x++) {
                                char suf[2];
                                if (!(x == (head.nx - 1))) {
                                        strcpy(suf,",");
                                }else{
                                        strcpy(suf,"");
                                }

                                if (g[z][y][x] >= head.max_voltage) {
                                        printf("%+2.3lf%s", g[z][y][x] - 2 * head.max_voltage , suf);
                                } else {
                                        printf("%+2.3lf%s", g[z][y][x], suf);
                                }

                        }
                        if (y == (head.ny - 1)){
                                printf("}\n");
                                continue;
                        }
                        printf("},\n");
                }
                if (z == (head.nz - 1)){
                        printf("}\n");
                        continue;
                }
                printf("},\n");
        }
        printf("-----------------------------------\n");
}

int is_electrode(PAHeader head, double point) {
	return point > head.max_voltage;
}

double average_potential(PAHeader head, double ***g) {
	int x, y, z;
	double average = 0;
	int electrode_count = 0;

	for (z = 0; z < head.nz; z++) {
		for (y = 0; y < head.ny; y++) {
			for (x = 0; x < head.nx; x++) {
				if (is_electrode(head, g[z][y][x])) {
					average += (g[z][y][x] - 2 * head.max_voltage);
					electrode_count++;
				}
			}
		}
	}

	average = average / electrode_count;
	printf("%d electrode points found, with an average potential of %lf\n", electrode_count, average);
	return average;
}


double Psi(double potential, double r) {
	return (unit_charge * potential)/T - ((e_mass * little_omega*(BigOmega - little_omega)*pow(r,2))/(2*T));
}

long double iterate(PAHeader head, long double ***g, double ***potential) {
	// x, y and z are coordinates, i is used to keep track of the number of points used for this average
	// Pn convention matches that used in section H.3 of the SIMION manual
	//
	// Attempting to solve e/T*potential -mw(W-w)r^2/(2T) where w/W = omega/Omega
	int x = 0; 
	int y = 0;
	int z = 0;
	int i = 0;
	
	long double avg_val = 0;
	long double max_change = 0;
	
	long double plasma_gamma = ((2 * e_mass * little_omega * (BigOmega - little_omega)) / (4 * 3.1415 * n0 * pow(unit_charge,2)));

	long double ***snapshot = NULL;
        snapshot = calloc(head.nz, sizeof(long double **));
        if (snapshot == NULL) {
                #ifdef DEBUG
                perror("Error allocating snapshot array (Z)");
                #endif
                return -1;
        }

        for (i = 0; i < head.nz; i++) {
                snapshot[i] = calloc(head.ny, sizeof(long double *));
                if (snapshot[i] == NULL) {
                        #ifdef DEBUG
                        perror("Error allocating snapshot array (Y)");
                        #endif
                        return -1;
                }
                int j = 0;
                for (j = 0; j < head.ny; j++) {
                        snapshot[i][j] = calloc(head.nx, sizeof(long double));
                        if (snapshot[i][j] == NULL) {
                                #ifdef DEBUG
                                perror("Error allocating snapshot array (X)");
                                #endif
                                return -1;
                        }
			int k = 0;
			for (k = 0; k < head.nx; k++) {
				snapshot[i][j][k]=g[i][j][k];
			}
		}
	}
	#ifdef DEBUG
	int maxX = -1;
	int maxY = -1;
	int maxZ = -1;
	#endif

	for (z = 0; z < head.nz; z++) {
		for (y = 0; y < head.ny; y++) {
			for (x = 0; x < head.nx; x++) {
				#ifdef DEBUG2
				printf("Processing point (%d, %d, %d)\n", x, y, z);	
				#endif
				avg_val = 0;
				i = 0;
				double rZ = fabs(z-head.nz/2);
				double rY = fabs(y-head.ny/2);
				double rX = fabs(x-head.nx/2);

				double r = sqrt(pow(rZ,2)+pow(rY,2)+pow(rX,2));
				if (head.nx > 1) {
					//At each point:
					//	Use snapshot as Psi1 and calculate RHS of eq 5
					//	Value at point = (average of surrounding snapshot points) - RHS Eq. 5
					//	Boundary conditions - easiest to define in terms of potential on cylinder wall?
					//	Define electrode potentials as constants above, test for equality when point is electrode (values in potential array)
					//

					if (g[z][y][x] != snapshot[z][y][x]) {
						fprintf(stderr, "Error: Snapshot inconsistent: g - %Lf	snapshot - %Lf\n", g[z][y][x], snapshot[z][y][x]);
						return -1;
					}
					if (potential[z][y][x] > head.max_voltage) {
						#ifdef DEBUG2
						printf("Electrode found at (%d,%d,%d)\n ", x, y, z);
						#endif
						if ((potential[z][y][x] - 2*head.max_voltage) == centralV) {
							g[z][y][x] = unit_charge / T * (endcapV - centralV) - (plasma_gamma) * pow(r,2) /4;
						} else if ((potential[z][y][x] - 2*head.max_voltage) == endcapV) {
							g[z][y][x] = unit_charge * centralV / T - (plasma_gamma) * pow(r,2) / 4;
						}
						#ifdef DEBUG2
						printf("Electrode - value set to %Lf", g[z][y][x]);
						#endif
						continue;
					}
					if (r==0 && x ==0) {
						g[z][y][x]=0;
						continue;
					}
					if (x == 0 && head.nx > 1) {
						#ifdef DEBUG3
						printf ("P0: (%d, %d, %d)\tP1: N/A  \tP2: (%d, %d, %d)\t\t", x, y, z, 1, y, z);
						#endif
						avg_val += snapshot[z][y][1] - exp(snapshot[z][y][1]); //P2
						i++;
						if (head.mirror & SIMION_MIRROR_X) { // Mirrored in X, so P2 also appears on the other side of the mirror
							avg_val += snapshot[z][y][1] - exp(snapshot[z][y][1]); //P2
							i++;
						}
					} else if (x == head.nx - 1 && (head.nx -1) > 0) {
						#ifdef DEBUG3
						printf ("P0: (%d, %d, %d)\tP1: (%d, %d, %d) \tP2: N/A  \t\t", x, y, z, x-1, y, z);
						#endif
						avg_val += snapshot[z][y][x-1] - exp(snapshot[z][y][x-1]); //P1
						i++;
					} else {
						#ifdef DEBUG3
						printf ("P0: (%d, %d, %d)\tP1: (%d, %d, %d) \tP2: (%d, %d, %d)\t\t", x, y, z, x-1, y, z, x+1, y, z);
						#endif
						avg_val += snapshot[z][y][x-1] - exp(snapshot[z][y][x-1]); //P1
						avg_val += snapshot[z][y][x+1] - exp(snapshot[z][y][x+1]); //P2
						i += 2;
					}
				}

				if (head.ny > 1) {
					if (y == 0 && head.ny > 1) {
						#ifdef DEBUG3
						printf ("P3: (%d, %d, %d) \tP4: N/A\t\t", x, y+1, z);
						#endif
						avg_val += snapshot[z][1][x] - exp(snapshot[z][1][x]); //P3
						i++;
						if (head.mirror & SIMION_MIRROR_Y) { //Mirrored in Y
							avg_val += snapshot[z][1][x] - exp(snapshot[z][1][x]); //P3
							i++;
						}
					} else if (y == head.ny - 1 && (head.ny - 1) > 0 ) {
						#ifdef DEBUG3
						printf ("P3: N/A  \tP4: (%d, %d, %d)\t\t", x, y-1, z);
						#endif
						avg_val += snapshot[z][(y-1)][x] - exp(snapshot[z][(y-1)][x]); //P4
						i++;
					} else {
						#ifdef DEBUG3
						printf ("P3: (%d, %d, %d) \tP4: (%d, %d, %d)\t\t", x, y+1, z, x, y-1, z);
						#endif
						avg_val += snapshot[z][(y+1)][x] - exp(snapshot[z][(y+1)][x]); //P3
						avg_val += snapshot[z][(y-1)][x] - exp(snapshot[z][(y-1)][x]); //P4
						i += 2;
					}
				}

				if (head.nz > 1) {
					if (z == 0) {
						#ifdef DEBUG3
						printf ("P5: (%d, %d, %d) \tP6: N/A\n", x, y, 1);
						#endif
						avg_val += snapshot[1][y][x] - exp(snapshot[1][y][x] -1); //P5
						i++;
						if (head.mirror & SIMION_MIRROR_Z) { //Mirrored in Z
							avg_val += snapshot[1][y][x] - exp(snapshot[1][y][x]); //P5
							i++;
						}
					} else if (z == head.nz - 1) {
						#ifdef DEBUG3
						printf ("P5: N/A  \tP6: (%d, %d, %d)\n", x, y, z-1);
						#endif
						avg_val += snapshot[z-1][y][x] - exp(snapshot[z-1][y][x]); //P6
						i++;
					} else {
						#ifdef DEBUG3
						printf ("P5: (%d, %d, %d) \tP6: (%d, %d, %d)\n", x, y, z+1, x, y, z-1);
						#endif
						avg_val += snapshot[z+1][y][x] - exp(snapshot[z+1][y][x]); //P5
						avg_val += snapshot[z-1][y][x] - exp(snapshot[z-1][y][x]); //P6
						i += 2;
					}
				}
	                        avg_val = avg_val / i;
	                        long double change = avg_val - snapshot[z][y][x];
	                        long double new_val = snapshot[z][y][x]+(change -(plasma_gamma));
 

	                        #ifdef DEBUG4
				if (x == head.nx/2) {
		                        printf("\nPoint (%d, %d):\n\tOld value: \t%+.8Lf\n\tAverages: \t%.8Lf\n", x, y, snapshot[z][y][x], avg_val);
					printf("\tChange: \t%.8Lf\n\tPotential: \t%+.8lf\n\tNew value: \t%.8Lf\n\ti:\t\t%d\n", change, potential[z][y][x], new_val, i);
				}
	                        #endif

	                        if (fabs(change) > max_change) {
	                                max_change = fabs(change);
					#ifdef DEBUG
					maxX = x;
					maxY = y;
					maxZ = z;
					#endif
	                        }
				#ifdef PAUSE
				usleep(PAUSE * 1000);
				#endif
	                        g[z][y][x] = new_val;

			}
		}
	}
	#ifdef DEBUG	
	printf("Max change: %.8Lf @ (%d,%d,%d). Was %.8LE, is now %.8LE\n", max_change, maxX, maxY, maxZ, snapshot[maxZ][maxY][maxX], g[maxZ][maxY][maxX]);
	#endif
        for (i = 0; i < head.nz; i++) {
                for (int j = 0; j < head.ny; j++) {
                        free(snapshot[i][j]);
		}
		free(snapshot[i]);
	}
	free(snapshot);
	return max_change;
}

int main(int argc, char *argv[]){
	fprintf(stderr,"FLTE Test program 1\nAttempt to solve for FLTE density profile based on *refined* PA");
	printf("%lf	%E\n%lf	%E\n", unit_charge, unit_charge, e_mass, e_mass);
	little_omega = 0.0000000000874619; //0.000000001 * unit_charge;
	BigOmega = unit_charge / (e_mass);
	T =  unit_charge * 100 * (centralV - endcapV);
	PAHeader pa_head;

	int iterations = 0;
	double goal = 0.005;
	long double del = 0;

	double ***potential = NULL;
	long double ***dist = NULL;
	
	if (argc == 2) {
		if ((potential = load_pa_file(argv[1], &pa_head, potential)) == NULL){
			perror("Error loading file");
			return EXIT_FAILURE;
		}
	} else if (argc == 3) {
		if (!(goal = atof(argv[2])) > 0) {
			fprintf(stderr, "Goal must be a positive number - not %+f\n", goal);
			return EXIT_FAILURE;
		}
		#ifdef DEBUG
		printf("Goal specified: %f\n", goal);
		#endif
		if ((potential = load_pa_file(argv[1], &pa_head, potential)) == NULL){
			perror("Error loading file");
			return EXIT_FAILURE;
		}
	} else {
		fprintf(stderr, "Usage: %s filename [goal]\nDefault goal is %f\n", argv[0], goal);
		return EXIT_FAILURE;
	}

	if (pa_head.symmetry == SIMION_CYLINDRICAL_SYMMETRY) {
		fprintf(stderr, "We know that cylindrical arrays don't work - enough already!\nAborting!");
		return EXIT_FAILURE;
	}

        dist = calloc(pa_head.nz, sizeof(long double **));
        if (dist == NULL) {
                #ifdef DEBUG
                perror("Error allocating dist array (Z)");
                #endif
                return EXIT_FAILURE;
        }

        int i = 0;
        for (i = 0; i < pa_head.nz; i++) {
                dist[i] = calloc(pa_head.ny, sizeof(long double *));
                if (dist[i] == NULL) {
                        #ifdef DEBUG
                        perror("Error allocating dist array (Y)");
                        #endif
                        return EXIT_FAILURE;
                }
                int j = 0;
                for (j = 0; j < pa_head.ny; j++) {
                        dist[i][j] = calloc(pa_head.nx, sizeof(long double));
                        if (dist[i][j] == NULL) {
                                #ifdef DEBUG
                                perror("Error allocating dist array (X)");
                                #endif
                                return EXIT_FAILURE;
                        }
			/*for (int k=0; k< pa_head.nx; k++) {
				double rZ = fabs(i-pa_head.nz/2);
				double rY = fabs(j-pa_head.ny/2);
				double rX = fabs(k-pa_head.nx/2);

				double r = sqrt(pow(rZ,2)+pow(rY,2)+pow(rX,2));
				dist[i][j][k] = (unit_charge / T) * potential[i][j][k] - (e_mass * little_omega * (BigOmega - little_omega) * pow(r,2))/(2*T);
			}*/
		}
	}	
	while ( !(iterations > 1000) ) {
		if ((del=iterate(pa_head, dist, potential)) <= goal && iterations > 10) {break;}
		iterations++;
		#ifdef DEBUG
		printf("Iteration: %d\nPsi @ centre: %+2.3Lf\n", iterations, dist[pa_head.nz/2][pa_head.ny/2][pa_head.nx/2]);
		#endif
	}

	if (iterations <= 1000) {
		printf("\nSolution after %d iterations:\n", iterations);
		display_grid(pa_head, potential);
	}else{
		fprintf(stderr, "\nError - solution did not converge sufficiently\n");
	}
	free_points(pa_head, potential);
	return EXIT_SUCCESS;
}
