#include "SimionPA.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

double*** load_pa_file(char *c, PAHeader *head, double ***points) {
	FILE *pa;
	errno = 0;
	if ((pa = fopen(c,"rb")) == NULL) {
		#ifdef DEBUG
		perror("Error loading file in load_pa_file");
		#endif
		return NULL;
	}
	head->mode = 0;
	head->symmetry = -1;
	head->max_voltage = -1;
	head->nx = 0;
	head->ny = 0;	
	head->nz = 0;
	head->ng = 0;
	head->mirror = -1;

	/*head->mode = (long) dgetc(pa) << 24;
	head->mode |= (long) dgetc(pa) << 16;
	head->mode |= (unsigned) (dgetc(pa) << 8);
	head->mode |= dgetc(pa);

	head->symmetry = (long) dgetc(pa) << 24;
	head->symmetry |= (long) dgetc(pa) << 16;
	head->symmetry |= (unsigned)(dgetc(pa) << 8);
	head->symmetry |= dgetc(pa);*/

	fread(&head->mode, sizeof(head->mode), 1, pa);
	fread(&head->symmetry, sizeof(head->symmetry), 1, pa);
	fread(&head->max_voltage, sizeof(head->max_voltage), 1, pa);
	fread(&head->nx, sizeof(head->nx), 1, pa);
	fread(&head->ny, sizeof(head->ny), 1, pa);
	fread(&head->nz, sizeof(head->nz), 1, pa);
	fread(&head->mirror, sizeof(head->mirror), 1, pa);
	head->ng = head->mirror >> 4;
	
	#ifdef DEBUG
	printf("File %s:\n\tMode: \t\t%d\n\tSymmetry: \t%d\n\tMax Voltage: \t%lf\n", c, head->mode, head->symmetry, head->max_voltage);
	printf("\tNx: \t\t%d\n\tNy: \t\t%d\n\tNz: \t\t%d\n\tNg: \t\t%d\n", head->nx, head->ny, head->nz, head->ng);
	printf("\tMirror: \t%d\n", head->mirror);
	if (head->symmetry == SIMION_PLANAR_SYMMETRY){
		printf("\tArray has planar symmetry\n");
	}
	if (head->symmetry == SIMION_CYLINDRICAL_SYMMETRY){
		printf("\tArray has cylindrical symmetry\n");
	}
	if (head->mirror & SIMION_MIRROR_X) {
		printf("\tArray mirrored in X axis\n");
	}
	if (head->mirror & SIMION_MIRROR_Y) {
		printf("\tArray mirrored in Y axis\n");
	}
	if (head->mirror & SIMION_MIRROR_Z) {
		printf("\tArray mirrored in Z axis\n");
	}
	if (head->mirror & SIMION_MAGNETIC) {
		printf("\tArray is a magnetic potential array\n");
	}
	
	#endif
	
	points = calloc(head->nz, sizeof(double **));
	if (points == NULL) {
		#ifdef DEBUG
		perror("Error allocating points array (Z)");
		#endif
		return NULL;
	}
	
	int i = 0;
	for (i = 0; i < head->nz; i++) {
		points[i] = calloc(head->ny, sizeof(double *));
		if (points[i] == NULL) {
			#ifdef DEBUG
			perror("Error allocating points array (Y)");
			#endif
			return NULL;
		}
		int j = 0;
		for (j = 0; j < head->ny; j++) {
			points[i][j] = calloc(head->nx, sizeof(double));
			if (points[i][j] == NULL) {
				#ifdef DEBUG
				perror("Error allocating points array (X)");
				#endif
				return NULL;
			}
			int k = fread(points[i][j], sizeof(double), head->nx, pa);
			if (!(k==head->nx)){
				#ifdef DEBUG
				fprintf(stderr, "Incomplete set of points read - read up to (%d,%d,%d)\n", head->nz, head-> ny, k);
				perror("");
				#endif
			}
		}
	}
	return points;
}

int save_pa_file(char *c, PAHeader *head, double ***points) {
	return 0;
}

void free_points(PAHeader head, double ***p){
	int i,j;
	for (i=0; i < head.nz; i++) {
		for (j=0; j < head.ny; j++) {
			free(p[i][j]);
		}
		free(p[i]);
	}
	free(p);
}
