/*****************************************************************************
 The definitions below are based on those printed in appendix F of the SIMION
 8.0 manual. While the definitions provided in that section  are not part of 
 the SL Libraries, this code may still fall under section 1(g) of the SIMION 
 License and should only be  used in conjunction with a valid license for 
 SIMION 8 or a license  for the SL Libraries
*****************************************************************************/
#ifndef SIMIONPA_H
#define SIMIONPA_H

#include <stdint.h>

#define SIMION_PLANAR_SYMMETRY 1
#define SIMION_CYLINDRICAL_SYMMETRY 0

/* SIMION_MIRROR_AXIS - Mirror in AXIS */
#define SIMION_MIRROR_X	1
#define SIMION_MIRROR_Y 2
#define SIMION_MIRROR_Z 4

/** 
 Magnetic potential array, rather than an 
 electrosotatic array 
**/
#define SIMION_MAGNETIC 8

struct simion_pa_header {
	int32_t mode ; /* = -1 */
	int32_t symmetry; /* = SIMION_CYLINDRICAL_SYMMETRY */
	double max_voltage; /* = 10000 */

	int32_t nx; /* = 50*/
	int32_t ny; /*= 50*/
	int32_t nz; /*= 50*/
/**
 SIMION works with magnetic potentials, which can be calculated from a measured
 gradient using the following defintions:

 Array Magnetic Potential = Array Gradient
 Magnetic Flux = Array MP * ng
 gauss = Array Gradient * ng

 ng is arbitrary and can be adjusted as required for an individual array
**/
	int32_t ng; /*= 100*/
/**
 The value stored in the mirror variable is defined as:
 (SIMION_MIRROR_X | SIMION_MIRROR_Y  | SIMION_MIRROR_Z | SIMION_MAGNETIC) 
  + (ng << 4)
 Where the values SIMION_MIRROR etc are either zero or the value defined above
**/
	int32_t mirror; /* = 0 */
};

typedef struct simion_pa_header PAHeader;

/**
 A PA file is simion_pa_header followed by double points[nz][ny][nx]
 Each point is the value of the potential at that point. For non-electrodes,
 the value at that point is stored as is. For electrodes, the value at that
 point is added to 2 * max_voltage
**/

double*** load_pa_file(char *c, PAHeader *head, double ***points);
int save_pa_file(char *c, PAHeader *head, double ***points); 
void free_points(PAHeader head, double ***p);
#endif
