#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define N 10

void display_grid(double *g){
	int y = 0;
	int x =0;
	
	printf("-----------------------------------\n");
	for (y=0; y<=(N-1); y++) {
		for (x=0; x<=(N-1); x++) {
			printf("\t%+2.3f", g[N*y+x]);
		}
		printf("\n");
	}
	printf("-----------------------------------\n\n");
}

double iterate(double *g, double orf) {
	// x and y are coordinates, i is used to keep track of the number of points used for this average
	// Pn convention matches that used in section H.3 of the SIMION manual
	int x = 0; 
	int y = 0;
	int i = 0;
	
	double avg_val = 0;
	double max_change = 0;

	for (y=0; y<=(N-1); y++) {
		for (x=0; x<=(N-1); x++) {
			// Preserve corner and centre points - treat them as electrodes
			if ( (x == 0 && y == 0) || (x == (N-1) && y == 0) || (x == 0 && y == (N-1)) || (x==(N-1) && y == (N-1)) ) { 
				#ifdef DEBUG
				printf("Skipping point (%d,%d)\n", x, y);
				#endif
				continue; 
			}
			#ifdef DEBUG
			printf("Processing point (%d,%d)\n", x, y);	
			#endif
			avg_val = 0;
			i = 0;
			if (x==0) {
				avg_val += g[N*y+1]; //P2
				i++;
			} else if (x==4) {
				avg_val += g[N*y+3]; //P1
				i++;
			} else {
				avg_val += g[N*y+(x-1)]; //P1
				avg_val += g[N*y+(x+1)]; //P2
				i += 2;
			}
			
			if (y==0) {
				avg_val += g[N+x]; //P3
				i++;
			} else if (y==(N-1)) {
				avg_val += g[N*(y-1)+x]; //P4
				i++;
			} else {
				avg_val += g[N*(y+1)+x]; //P3
				avg_val += g[N*(y-1)+x]; //P4
				i += 2;
			}
                        avg_val = avg_val / i;
                        double change = avg_val - g[N*y+x];
                        double new_val = g[N*y+x]+(change) * (1 + orf);

                        #ifdef DEBUG2
                        printf("\nPoint (%d, %d):\n\tOld value: \t%+.3f\n\tAverages: \t%.3f\n", x, y, g[N*y+x], avg_val);
                        printf("\tChange: \t%.3f\n\tNew value: \t%.3f\n\ti:\t\t%d\n", change, new_val, i);
                        #endif

                        if (fabs(change) > max_change) {
                                max_change = fabs(change);
                        }
                        g[N*y+x] = new_val;
	
		}
	}
	#ifdef DEBUG	
	printf("Max change: %.3f\n", max_change);
	#endif
	return max_change;
}

int main(int argc, char *argv[]){
	printf("Simple Over-Relaxation test\n\nIterate over an %dx%d grid and print values for each iteration\n", N, N);

	if (!(argc==2)){
		printf("Usage: %s <orf>\n 0 < orf < 1\n", argv[0]);
		return EXIT_FAILURE;
	}
	
	double orf = atof(argv[1]);
	if ( orf < 0 || orf > 1) {
		printf("Usage: %s <orf>\n 0 < orf < 1\n", argv[0]);
		return EXIT_FAILURE;
	}

	//double grid[(N*N)]={5,0,0,0,5,0,0,0,0,0,0,0,-1,0,0,0,0,0,0,0,5,0,0,0,5};
	double *grid = calloc(N*N, sizeof(double));
	
	if (grid == NULL) {
		printf("Unable to allocate array (Requested a %dx%d array - %ld bytes)", N, N, N*N*sizeof(double));
		return EXIT_FAILURE;
	}	
	
	grid[0] = +5;
	grid[N-1] = +5;
	grid[N*(N-1)] = +5;
	grid[N*N-1] = +5;

	display_grid(grid);
	
	int iterations = 0;

	while (iterate(grid, orf) > 0.01) {
		iterations++;
		printf("Iteration: %d\n", iterations);
		display_grid(grid);
	}
	display_grid(grid);
	free(grid);
	return EXIT_SUCCESS;
}
